import base64
import json
from binascii import hexlify, unhexlify
from typing import Any
from pydantic import BaseModel
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
from cryptography.hazmat.backends import default_backend


class KeyParams(BaseModel):
    nonce: str
    tag: str


class Params(BaseModel):
    nonce: str
    tag: str


class Slot(BaseModel):
    type: int
    salt: str
    n: int
    r: int
    p: int
    key_params: KeyParams
    key: str


class Header(BaseModel):
    slots: list[Slot]
    params: Params

    
class Vault(BaseModel):
    version: int
    header: Header
    db: Any  # This will be the parsed JSON content


class VaultEncrypted(BaseModel):
    version: int
    header: Header
    db: str
    
    def find_master_key(self, pwd: str) -> bytes:
        """
        Uses the password to decrypt the master key from the vault and returns the master key's bytes.
        
        Args:
            pwd: The password to use for decryption
            
        Returns:
            The decrypted master key as bytes
            
        Raises:
            ValueError: If no master key is found
        """
        key = None
        master_key = None
        
        for slot in self.header.slots:
            # Ignore slots that aren't using the password type
            if slot.type != 1:
                continue
                
            salt = unhexlify(slot.salt)
            
            # Create a key using the slot values and provided password
            kdf = Scrypt(
                salt=salt,
                length=32,
                n=slot.n,
                r=slot.r,
                p=slot.p,
                backend=default_backend()
            )
            key = kdf.derive(pwd.encode())
            
            nonce = unhexlify(slot.key_params.nonce)
            tag = unhexlify(slot.key_params.tag)
            slot_key = unhexlify(slot.key)
            
            # Combine slot key and tag for decryption
            key_data = slot_key + tag
            
            # Attempt to decrypt the master key
            try:
                aesgcm = AESGCM(key)
                master_key = aesgcm.decrypt(nonce, key_data, None)
                # If decryption succeeds, break out of the loop
                break
            except Exception:
                # Continue to the next slot if decryption fails
                continue
        
        if master_key is None or len(master_key) == 0:
            raise ValueError("no master key found")
            
        return master_key
    
    def decrypt_contents(self, master_key: bytes) -> bytes:
        """
        Uses the master key to decrypt the vault's contents and returns the content's bytes.
        
        Args:
            master_key: The master key as bytes
            
        Returns:
            The decrypted contents as bytes
            
        Raises:
            Exception: If decryption fails
        """
        db_str = self.db
        params = self.header.params
        
        nonce = unhexlify(params.nonce)
        tag = unhexlify(params.tag)
        db_data = base64.b64decode(db_str)
        
        # Combine database data and tag for decryption
        database = db_data + tag
        
        # Attempt to decrypt the vault content
        try:
            aesgcm = AESGCM(master_key)
            content = aesgcm.decrypt(nonce, database, None)
            return content
        except Exception as e:
            raise e
    
    def decrypt_vault(self, master_key: bytes) -> Vault:
        """
        Decrypts the vault's contents and returns a plaintext version of the vault.
        
        Args:
            master_key: The master key as bytes
            
        Returns:
            A Vault instance with decrypted content
            
        Raises:
            Exception: If decryption or JSON parsing fails
        """
        content = self.decrypt_contents(master_key)
        
        # Parse the JSON content
        db_content = json.loads(content.decode('utf-8'))
        
        # Create a plaintext vault with decrypted content
        vault_data_plain = Vault(
            version=self.version,
            header=self.header,
            db=db_content
        )
        
        return vault_data_plain




import sys
import pathlib
import json
import pprint

import pyotp

from py_avdu.encrypted_classes import KeyParams, Params, Header, Slot, VaultEncrypted
from py_avdu.decrypted_classes import Db


def main(args = sys.argv[1:]):
    vault_path, pwd = args

    sys.argv.clear()
    args.clear()

    vault_dict = json.loads(pathlib.Path(vault_path).read_text())

    encrypted = VaultEncrypted(**vault_dict)

    master_key = encrypted.find_master_key(pwd)

    del pwd

    decrypted = encrypted.decrypt_vault(master_key)

    del master_key

    db_plain = Db(**decrypted.db)

    del decrypted

    for entry in db_plain.entries:
        info = entry.info
        totp = pyotp.TOTP(info.secret)
        print(f'{entry.issuer}, {entry.name}: {totp.now()} ')

    del entry, info, db_plain, totp



if __name__ == '__main__':
    main()


from typing import Optional
from pydantic import BaseModel


class Info(BaseModel):
    secret: str
    algo: str
    digits: int
    period: Optional[int] = None
    counter: Optional[int] = None
    pin: Optional[str] = None


class Group(BaseModel):
    uuid: str
    name: str


class Entry(BaseModel):
    type: str
    uuid: str
    name: str
    issuer: str
    note: str
    icon: Optional[str] = None
    icon_mime: Optional[str] = None
    icon_hash: Optional[str] = None
    favorite: bool
    info: Info
    groups: list[str]


class Db(BaseModel):
    version: int
    entries: list[Entry]
    groups: list[Group]